const _checkCondition = (idArr, query, pkColumnName, isChild) => {
  let deleteQuery = query.replace("andCondition", `and ${isChild ? 'ct.'+pkColumnName : pkColumnName} not in (${idArr.join(", ")})`);
  if (!idArr.length) {
    deleteQuery = query.replace("andCondition", "");
  }
  return deleteQuery;
};

/**
 * Represents Postgres Service to perform action on postgreSQL database
 */
class PostgresService {
  /**
   * @constructor
   * @param {LoggerService} loggerService - Logger wrapper,
   * @param {Pool} pool - postgre DB connection
   */
  constructor({ loggerService, pool }) {
    this._loggerService = loggerService;
    this._pool = pool;
    this._loggerService.debug(`PostgresService initialized`);
  }

  /**
   * @description - Execute postgreSQL database query
   *
   * @async
   * @function query
   * @param {string} literal - Database query literal
   * @param {array} params - Dynamic value in query string
   */
  async query(literal, params = "") {
    try {
      return await this._pool.query(literal, params);
    } catch (error) {
      this._errorHandler(error);
    }
  }
  
    /**
   * @description - Execute postgreSQL database multiple query in transaction i.e rollback if error occur
   *
   * @async
   * @function transaction
   * @param {string} multipleInsertQueryString - Database query literal for multiple insert
   * @param {string} updateQueryStringArr - Database query literal for update
   * @param {string} childDeleteQueryString - Database query literal for parent records delete
   * @param {string} parentDeleteQueryString - Database query literal for child records delete
   * @param {string} pkColumnName - Primary key column name of table
   * @param {uploadService} uploadService - Aws s3 service wrapper
   */
  async transaction(multipleInsertQueryString, updateQueryStringArr, parentDeleteQueryString, childDeleteQueryString, pkColumnName, uploadService) {
    // note: we don't try/catch this because if connecting throws an exception
    // we don't need to dispose of the client (it will be undefined)
    const client = await this._pool.connect();
    const promiseArr = [];
    const idArr = [];
    try {
      await client.query("BEGIN");
      multipleInsertQueryString && promiseArr.push(client.query(multipleInsertQueryString));

      // todo check for bulk update
      for (let i = 0; i < updateQueryStringArr.length; i++) {
        promiseArr.push(client.query(updateQueryStringArr[i].queryString, updateQueryStringArr[i].valuesArr));
      }

      return await Promise.all(promiseArr).then(async values => {
        values.map(result => {
          result.rows.length &&
            result.rows.forEach(obj => {
              idArr.push(obj[pkColumnName]);
            });
        });

        let deletedRecords = [];
        const newPromiseArr = [];
        if (childDeleteQueryString.length) {
          childDeleteQueryString.forEach(async query => {
            const childDeleteQuery = _checkCondition(idArr, query, pkColumnName, true);
            newPromiseArr.push(client.query(childDeleteQuery));
          });
          await Promise.all(newPromiseArr).then(async res => {
            res.map(result => {
              if (result.rows.length) {
                deletedRecords = deletedRecords.concat(result.rows);
              }
            });
          }).catch(async error => {
            await client.query("ROLLBACK");
            throw error;
          });
        }

        const parentDeleteQuery = _checkCondition(idArr, parentDeleteQueryString, pkColumnName);
        const { rows } = await client.query(parentDeleteQuery);
        deletedRecords = deletedRecords.concat(rows);
        const fileArr = [];
        deletedRecords.length && deletedRecords.filter(obj => {
          if (obj.attachment_path || obj.attachment_file_location) {
            fileArr.push({
              "Key": obj.attachment_path || obj.attachment_file_location
            });
          }
        });

        if (fileArr && fileArr.length) {
          await uploadService.deleteFileFromS3(fileArr);
        }
        await client.query("COMMIT");

      }).catch(async error => {
        await client.query("ROLLBACK");
        throw error;
      });
    } catch (e) {
      await client.query("ROLLBACK");
      throw e.message || e;
    }
     finally {
      client.release();
    }
  }

  /**
   * @description - Handle errors thrown by any method
   * @private
   * @function _errorHandler
   * @param {object} error - Error with message thrown by any method
   */
  _errorHandler(error) {
    this._loggerService.error("An error occurred in postgres db service", error);
    throw error;
  }
}

module.exports = PostgresService;
